"""Provides utilties used to manage Nawah App serving"""

import asyncio
import datetime
import logging
import traceback
from functools import partial
from typing import TYPE_CHECKING

from nawah.config import Config

if TYPE_CHECKING:
    from asyncio import Future

    from nawah.classes import Job

logger = logging.getLogger("nawah")


def _populate_routes():
    get_routes = []
    post_routes = []
    for module_name, module in Config.modules.items():
        for func_name, func in module.funcs.items():
            if func.get_func:
                for get_args_set in func.query_attrs or [{}]:
                    if get_args_set:
                        get_args = f'/{{{"}/{".join(list(get_args_set.keys()))}}}'
                    else:
                        get_args = ""

                    get_routes.append(f"/{module_name}/{func_name}{get_args}")

            func.post_func = True
            if func.post_func:
                post_routes.append(f"/{module_name}/{func_name}")

    return (get_routes, post_routes)


async def _check_sessions(session_ids: list[str], /):
    for session_id in session_ids:
        try:
            session = Config.sys.sessions[session_id]
            if "last_call" not in session["conn"]:
                continue
            if datetime.datetime.utcnow() > (
                session["conn"]["last_call"]
                + datetime.timedelta(seconds=Config.conn_timeout)
            ):
                logger.debug(
                    "Session #'%s' with REMOTE_ADDR '%s' HTTP_USER_AGENT: '%s' is idle. "
                    "Closing",
                    session["conn"]["id"],
                    session["conn"]["REMOTE_ADDR"],
                    session["conn"]["HTTP_USER_AGENT"],
                )
                asyncio.create_task(_close_session(session_id))
        except Exception:
            logger.error("An error occurred. Details: %s", traceback.format_exc())


async def _close_session(session_id: str, /):
    # [TODO] Check necissity to implement session_lock for deleteing env object from sessions

    if session_id not in Config.sys.sessions:
        logger.debug("Skipped closing session #'%s'", session_id)
        return

    # Remove id from Env dict to avoid having _close_session be called from _websocket_handler
    # upon closing the websocket connection
    del Config.sys.sessions[session_id]["conn"]["id"]

    logger.debug(
        "Closing data connection for session #'%s'",
        session_id,
    )
    Config.sys.sessions[session_id]["conn"]["data"].close()

    logger.debug("Done closing data connection")
    logger.debug(
        "Websocket connection status: %s",
        not Config.sys.sessions[session_id]["conn"]["ws"].closed,
    )

    if not Config.sys.sessions[session_id]["conn"]["ws"].closed:
        await Config.sys.sessions[session_id]["conn"]["ws"].close()
    logger.debug("Websocket connection for session #'%s' closed", session_id)

    try:
        del Config.sys.sessions[session_id]
    except KeyError:
        logger.error("Failed to delete session #%s", session_id)


async def _check_ips_quota(ips: list[str], /):
    for ip in ips:
        ip_quota = Config.sys.ip_quota[ip]
        try:
            if (datetime.datetime.utcnow() - ip_quota["last_check"]).seconds > 59:
                logger.debug(
                    "IP '%s' with quota '%s' is idle. Cleaning-up",
                    ip,
                    ip_quota["counter"],
                )
                del Config.sys.ip_quota[ip]
        except Exception:
            logger.error("An error occurred. Details: %s", traceback.format_exc())


async def _execute_jobs():
    current_time = datetime.datetime.utcnow().isoformat()[:16]
    for job_name, job in Config.jobs.items():
        try:
            logger.debug("Checking: %s", job_name)
            if job.disabled:
                logger.debug("-Job is disabled. Skipping")
                continue
            # Check if job is scheduled for current_time
            if current_time >= job.next_time:
                logger.debug("-Job is due, running!")
                # Update job next_time
                job.next_time = datetime.datetime.fromtimestamp(
                    job.cron_schedule.get_next(), datetime.timezone.utc
                ).isoformat()[:16]

                job_task = asyncio.create_task(job.job(session=Config.sys.session))
                job_task.add_done_callback(
                    partial(_job_callback, job_name=job_name, job=job)
                )

            else:
                logger.debug("-Not yet due")
        except Exception:
            logger.error("An error occurred. Details: %s", traceback.format_exc())


def _job_callback(job_task: "Future", job_name: str, job: "Job"):
    try:
        job_task.result()
        logger.debug("-Job '%s' is done", job_name)
    except Exception as e:
        logger.error("Job '%s' has failed with exception", job_name)
        logger.error("Exception details:")
        logger.error(e)
        if job.prevent_disable:
            logger.warning("-Detected job prevent_disable. Skipping disabling job")
        else:
            logger.warning("-Disabling job")
            job.disabled = True



"""Provides 'delete_file' Base Function callable"""

from typing import TYPE_CHECKING

from nawah.classes import Query
from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import InvalidAttrException
from nawah.utils import call

from .exceptions import (InvalidDeleteFileDocAttrException,
                         InvalidDeleteFileIndexException,
                         InvalidDeleteFileIndexValueException,
                         InvalidDeleteFileMismatchNameException,
                         InvalidDocException, UtilityModuleDataCallException)

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahSession, Results


async def delete_file(
    *,
    module_name: str,
    session: "NawahSession",
    query: "Query",
    doc: "NawahDoc",
) -> "Results":
    """Removes file from existing doc in module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="delete_file"
        )

    # [TODO] Allow use dot-notated attr path in attr query attr
    file_attr_name = query["attr:$eq"][0]
    if (
        file_attr_name not in module.attrs
        or module.attrs[file_attr_name].type != AttrType.LIST
        or not module.attrs[file_attr_name].args["list"][0].type != AttrType.FILE
    ):
        raise InvalidAttrException(
            attr_name=file_attr_name,
            attr_type=module.attrs[file_attr_name].type
            if file_attr_name in module.attrs
            else None,
            val_type=type(doc[file_attr_name]),
        )

    read_results = await call(
        "base/read",
        module_name=module_name,
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
    )

    if not read_results["args"]["count"]:
        raise InvalidDocException(doc_id=query["_id:$eq"][0])

    doc = read_results["args"]["docs"][0]

    if file_attr_name not in doc or not doc[file_attr_name]:
        raise InvalidDeleteFileDocAttrException(
            doc_id=query["_id:$eq"][0], attr_name=file_attr_name
        )

    if query["index:$eq"][0] not in range(len(doc[file_attr_name])):
        raise InvalidDeleteFileIndexException(
            doc_id=query["_id:$eq"][0],
            attr_name=file_attr_name,
            index=query["index:$eq"][0],
        )

    if (
        not isinstance(doc[file_attr_name][query["index:$eq"][0]], dict)
        or "name" not in doc[file_attr_name][query["index:$eq"][0]]
    ):
        raise InvalidDeleteFileIndexValueException(
            doc_id=query["_id:$eq"][0],
            attr_name=file_attr_name,
            index=query["index:$eq"][0],
            index_val_type=type(doc[file_attr_name][query["index:$eq"][0]]),
        )

    if doc[file_attr_name][query["index:$eq"][0]]["name"] != query["name:$eq"][0]:
        raise InvalidDeleteFileMismatchNameException(
            doc_id=query["_id:$eq"][0],
            attr_name=file_attr_name,
            index=query["index:$eq"][0],
            query_file_name=query["name:$eq"][0],
            index_file_name=doc[file_attr_name][query["index:$eq"][0]]["name"],
        )

    update_results = await call(
        "base/update",
        module_name=module_name,
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
        doc={
            file_attr_name: {"$del_val": [doc[file_attr_name][query["index:$eq"][0]]]}
        },
    )

    return update_results



"""Provides """

import datetime
from typing import TYPE_CHECKING, Literal, MutableMapping, MutableSequence

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahSession, Results, ResultsArgs


def mock_results(
    *, status: int, msg: str = None, args: "ResultsArgs" = None
) -> "Results":
    """Returns dict structured as call results"""

    return {
        "status": status,
        "msg": msg or "Mock results",
        "args": args or {},
    }


def mock_session(
    *, user: MutableMapping = None, groups: MutableSequence[str] = None
) -> "NawahSession":
    """Returns dict structured as 'session' of 'Env' dict. If no value provided for 'user', value
    would be return of 'mock_user()'. If no value provided for 'groups', value would be empty
    list"""

    return {
        "user": user or {},
        "groups": groups or [],
        "host_add": "127.0.0.1",
        "user_agent": "Nawah Framework Testing Tools",
        "expiry": datetime.datetime.utcnow().isoformat(),
        "token_hash": "__token",
        "create_time": datetime.datetime.utcnow().isoformat(),
        "client_app": "__test",
        "args": {},
    }


def mock_user(  # pylint: disable=too-many-arguments
    *,
    _id: ObjectId = None,
    name: MutableMapping[str, str] = None,
    locale: str = None,
    create_time: str = None,
    login_time: str = None,
    groups: MutableSequence[str] = None,
    privileges: MutableMapping[str, MutableSequence[str]] = None,
    status: Literal["active", "banned", "deleted", "disabled_password"] = None,
    **kwargs
) -> MutableMapping:
    """Thin wrapper for 'mock_doc'. Returns dict structured as 'user' doc"""

    return mock_doc(
        _id=_id,
        name=name or {locale: "__name" for locale in Config.locales},
        locale=locale or Config.locale,
        create_time=create_time or datetime.datetime.utcnow().isoformat(),
        login_time=login_time or datetime.datetime.utcnow().isoformat(),
        groups=groups or ["f00000000000000000000013"],
        privileges=privileges or {},
        status=status,
        **kwargs,
    )


def mock_doc(_id: ObjectId = None, **kwargs):
    """Returns dict structured as doc"""

    return {
        "_id": _id or ObjectId(),
        **kwargs,
    }



"""Provides utilities related to Attr Type"""

import copy
import datetime
import logging
import math
import random
import re
from typing import Any, MutableMapping, MutableSequence, cast

from bson import ObjectId

from nawah.classes import Attr
from nawah.config import Config
from nawah.enums import AttrType, NawahValues
from nawah.exceptions import InvalidAttrTypeException

from ._shared import ATTRS_TYPES_ARGS
from ._validate import validate_type

logger = logging.getLogger("nawah")


def extract_attr(*, attrs: MutableMapping[str, "Attr"], path: str) -> "Attr":
    """Extracts attr using its 'path' from 'attrs'. Attr path considers different Attr Types. For
    instance, to get child attr of Attr Type LIST, UNION, use index of child attr, to get 'key',
    'val' attrs of Attr Type KV_DICT use the arg name, to get child attr of Attr Type TYPED_DICT
    use key from 'dict' arg"""

    scope: MutableMapping[str, "Attr"] | MutableSequence["Attr"] = attrs
    path_part: int | str
    path_parts = path.split(".")
    _iter = enumerate(path_parts)
    for _, path_part in _iter:
        if re.match("^[0-9]+$", cast(str, path_part)):
            path_part = int(path_part)
            path_part_check = path_part in range(len(scope))
        else:
            path_part_check = path_part in scope

        if not path_part_check:
            raise ValueError(f"Attr '{path_part}' does not exist in scope '{scope}'")

        scope_attr = cast(MutableMapping[str, "Attr"], scope)[cast(str, path_part)]

        while scope_attr.type in [AttrType.LIST, AttrType.UNION]:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            arg = "list" if scope_attr.type == AttrType.LIST else "union"
            scope_attr = scope_attr.args[arg][int(path_part)]

        while scope_attr.type == AttrType.KV_DICT:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            scope_attr = scope_attr.args[path_part]

        while scope_attr.type == AttrType.TYPED_DICT:
            try:
                _, path_part = next(_iter)
            except StopIteration:
                return scope_attr

            scope_attr = scope_attr.args["dict"][path_part]

        match scope_attr.type:
            case AttrType.LIST:
                scope = scope_attr.args["list"]
            case AttrType.UNION:
                scope = scope_attr.args["union"]
            case AttrType.KV_DICT:
                scope = scope_attr.args["val"]
            case AttrType.TYPED_DICT:
                scope = scope_attr.args["dict"]

    return scope_attr


def encode_attr_type(*, attr_type: "Attr") -> MutableMapping[str, Any]:
    """Encodes Attr object into dict"""

    encoded_attr_type: MutableMapping[str, Any] = {
        "type": attr_type.type,
        "args": copy.deepcopy(attr_type.args),
        "allow_none": attr_type.default != NawahValues.NONE_VALUE,
        "default": attr_type.default
        if attr_type.default != NawahValues.NONE_VALUE
        else None,
    }
    # [DOC] Process args of type Attr
    if attr_type.type == AttrType.LIST:
        for i in range(len(attr_type.args["list"])):
            encoded_attr_type["args"]["list"][i] = encode_attr_type(
                attr_type=attr_type.args["list"][i]
            )
    elif attr_type.type == AttrType.TYPED_DICT:
        for dict_attr in attr_type.args["dict"].keys():
            encoded_attr_type["args"]["dict"][dict_attr] = encode_attr_type(
                attr_type=attr_type.args["dict"][dict_attr]
            )
    elif attr_type.type == AttrType.KV_DICT:
        encoded_attr_type["args"]["key"] = encode_attr_type(
            attr_type=attr_type.args["key"]
        )
        encoded_attr_type["args"]["val"] = encode_attr_type(
            attr_type=attr_type.args["val"]
        )
    elif attr_type.type == AttrType.UNION:
        for i in range(len(attr_type.args["union"])):
            encoded_attr_type["args"]["union"][i] = encode_attr_type(
                attr_type=attr_type.args["union"][i]
            )
    elif attr_type.type == AttrType.TYPE:
        if callable(encoded_attr_type["args"]["type"]):
            raise Exception("Attr Type TYPE with callable 'type' can't be encoded.")

        del encoded_attr_type["args"]["func"]

    return encoded_attr_type


def decode_attr_type(*, encoded_attr_type: MutableMapping[str, Any]) -> "Attr":
    """Converts Attr Type decoded into dict, back to Attr object"""
    # Fail-safe checks
    if encoded_attr_type["type"] not in ATTRS_TYPES_ARGS:
        raise InvalidAttrTypeException(attr_type=encoded_attr_type["type"])

    # Process args of type Attr
    if encoded_attr_type["type"] == "LIST":
        for i, _ in enumerate(encoded_attr_type["args"]["list"]):
            encoded_attr_type["args"]["list"][i] = decode_attr_type(
                encoded_attr_type=encoded_attr_type["args"]["list"][i]
            )
    elif encoded_attr_type["type"] == "TYPED_DICT":
        for dict_attr in encoded_attr_type["args"]["dict"].keys():
            encoded_attr_type["args"]["dict"][dict_attr] = decode_attr_type(
                encoded_attr_type=encoded_attr_type["args"]["dict"][dict_attr]
            )
    elif encoded_attr_type["type"] == "KV_DICT":
        encoded_attr_type["args"]["key"] = decode_attr_type(
            encoded_attr_type=encoded_attr_type["args"]["key"]
        )
        encoded_attr_type["args"]["val"] = decode_attr_type(
            encoded_attr_type=encoded_attr_type["args"]["val"]
        )
    if encoded_attr_type["type"] == "UNION":
        for i in range(len(encoded_attr_type["args"]["union"])):
            encoded_attr_type["args"]["union"][i] = decode_attr_type(
                encoded_attr_type=encoded_attr_type["args"]["union"][i]
            )
    # Generate dynamic Attr using Attr controller
    attr_type = getattr(Attr, encoded_attr_type["type"])(**encoded_attr_type["args"])

    # Validate Attr Type for possible Attr Type TYPE presence, to set Attr._args[func] value
    validate_type(attr_type=attr_type)

    return attr_type


def generate_attr_val(*, attr_type: Attr) -> Any:
    attr_val: Any

    if attr_type.type == AttrType.ANY:
        return "__any"

    if attr_type.type == AttrType.BOOL:
        attr_val = random.choice([True, False])
        return attr_val

    if attr_type.type == AttrType.DATE:
        if attr_type.args["ranges"]:
            datetime_range = attr_type.args["ranges"][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ["+", "-"] and datetime_range[1][0] in [
                "+",
                "-",
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == "d":
                    datetime_range_delta = {"days": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "w":
                    datetime_range_delta = {"weeks": int(datetime_range[0][:-1])}
                attr_val = (
                    (
                        datetime.datetime.utcnow()
                        + datetime.timedelta(**datetime_range_delta)
                    )
                    .isoformat()
                    .split("T")[0]
                )
            else:
                if datetime_range[0][0] not in ["+", "-"]:
                    attr_val = datetime_range[0]
                else:
                    attr_val = (
                        (
                            datetime.datetime.fromisoformat(datetime_range[1])
                            - datetime.timedelta(days=1)
                        )
                        .isoformat()
                        .split("T")[0]
                    )
        else:
            attr_val = datetime.datetime.utcnow().isoformat().split("T")[0]
        return attr_val

    if attr_type.type == AttrType.DATETIME:
        if attr_type.args["ranges"]:
            datetime_range = attr_type.args["ranges"][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ["+", "-"] and datetime_range[1][0] in [
                "+",
                "-",
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == "d":
                    datetime_range_delta = {"days": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "s":
                    datetime_range_delta = {"seconds": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "m":
                    datetime_range_delta = {"minutes": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "h":
                    datetime_range_delta = {"hours": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "w":
                    datetime_range_delta = {"weeks": int(datetime_range[0][:-1])}
                attr_val = (
                    datetime.datetime.utcnow()
                    + datetime.timedelta(**datetime_range_delta)
                ).isoformat()
            else:
                if datetime_range[0][0] not in ["+", "-"]:
                    attr_val = datetime_range[0]
                else:
                    attr_val = (
                        datetime.datetime.fromisoformat(datetime_range[1])
                        - datetime.timedelta(days=1)
                    ).isoformat()
        else:
            attr_val = datetime.datetime.utcnow().isoformat()
        return attr_val

    if attr_type.type == AttrType.KV_DICT:
        attr_val = {}
        if attr_type.args["req"]:
            attr_val = {
                generate_attr_val(
                    attr_type=Attr.LITERAL(literal=[req])
                ): generate_attr_val(attr_type=attr_type.args["val"])
                for req in attr_type.args["req"]
            }
        if attr_type.args["len_range"]:
            for _ in range(attr_type.args["len_range"][0]):
                attr_val[
                    generate_attr_val(attr_type=attr_type.args["key"])
                ] = generate_attr_val(attr_type=attr_type.args["val"])
            return attr_val

        return attr_val

    if attr_type.type == AttrType.TYPED_DICT:
        attr_val = {
            child_attr: generate_attr_val(attr_type=attr_type.args["dict"][child_attr])
            for child_attr in attr_type.args["dict"].keys()
        }
        return attr_val

    if attr_type.type == AttrType.EMAIL:
        attr_val = f"some-{math.ceil(random.random() * 10000)}@mail.provider.com"
        if attr_type.args["allowed_domains"]:
            if attr_type.args["strict_matching"]:
                domain = "mail.provider.com"
            else:
                domain = "provider.com"
            attr_val = attr_val.replace(
                domain, random.choice(attr_type.args["allowed_domains"])
            )
        return attr_val

    if attr_type.type == AttrType.FILE:
        attr_file_type = "text/plain"
        attr_file_extension = "txt"
        if attr_type.args["types"]:
            for file_type in attr_type.args["types"]:
                if "/" in file_type:
                    attr_file_type = file_type
                if "*." in file_type:
                    attr_file_extension = file_type.replace("*.", "")
        file_name = f"__file-{math.ceil(random.random() * 10000)}.{attr_file_extension}"
        return {
            "name": file_name,
            "lastModified": 100000,
            "type": attr_file_type,
            "size": 6,
            "content": b"__file",
        }

    if attr_type.type == AttrType.FLOAT:
        if attr_type.args["ranges"]:
            attr_val = random.choice(
                range(
                    math.ceil(attr_type.args["ranges"][0][0]),
                    math.floor(attr_type.args["ranges"][0][1]),
                )
            )
            if (
                attr_val != attr_type.args["ranges"][0][0]
                and (attr_val - 0.01) != attr_type.args["ranges"][0][0]
            ):
                attr_val -= 0.01
            elif (attr_val + 0.01) < attr_type.args["ranges"][0][1]:
                attr_val += 0.01
            else:
                attr_val = float(attr_val)
        else:
            attr_val = random.random() * 10000
        return attr_val

    if attr_type.type == AttrType.GEO_POINT:
        return {
            "type": "Point",
            "coordinates": [
                math.ceil(random.random() * 100000) / 1000,
                math.ceil(random.random() * 100000) / 1000,
            ],
        }

    if attr_type.type == AttrType.ID:
        return ObjectId()

    if attr_type.type == AttrType.INT:
        if attr_type.args["ranges"]:
            attr_val = random.choice(
                range(attr_type.args["ranges"][0][0], attr_type.args["ranges"][0][1])
            )
        else:
            attr_val = math.ceil(random.random() * 10000)
        return attr_val

    if attr_type.type == AttrType.IP:
        return "127.0.0.1"

    if attr_type.type == AttrType.LIST:
        return [
            generate_attr_val(attr_type=random.choice(attr_type.args["list"]))
            for _ in range(
                attr_type.args["len_range"][0] if attr_type.args["len_range"] else 0
            )
        ]

    if attr_type.type == AttrType.LOCALE:
        return {
            locale: f"__locale-{math.ceil(random.random() * 10000)}"
            for locale in Config.locales
        }

    if attr_type.type == AttrType.LOCALES:
        return Config.locale

    if attr_type.type == AttrType.PHONE:
        attr_phone_code = "000"
        if attr_type.args["codes"]:
            attr_phone_code = random.choice(attr_type.args["codes"])
        return f"+{attr_phone_code}{math.ceil(random.random() * 10000)}"

    if attr_type.type == AttrType.STR:
        if attr_type.args["pattern"]:
            logger.warning(
                "Generator for Attr Type STR can't handle patterns. Ignoring."
            )
        return f"__str-{math.ceil(random.random() * 10000)}"

    if attr_type.type == AttrType.TIME:
        if attr_type.args["ranges"]:
            datetime_range = attr_type.args["ranges"][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ["+", "-"] and datetime_range[1][0] in [
                "+",
                "-",
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == "s":
                    datetime_range_delta = {"seconds": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "m":
                    datetime_range_delta = {"minutes": int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == "h":
                    datetime_range_delta = {"hours": int(datetime_range[0][:-1])}
                attr_val = (
                    (
                        datetime.datetime.utcnow()
                        + datetime.timedelta(**datetime_range_delta)
                    )
                    .isoformat()
                    .split("T")[1]
                )
            else:
                if datetime_range[0][0] not in ["+", "-"]:
                    attr_val = datetime_range[0]
                else:
                    # [REF]: https://stackoverflow.com/a/656394/2393762
                    attr_val = (
                        (
                            datetime.datetime.combine(
                                datetime.date.today(),
                                datetime.time.fromisoformat(datetime_range[1]),
                            )
                            - datetime.timedelta(minutes=1)
                        )
                        .isoformat()
                        .split("T")[1]
                    )
        else:
            attr_val = datetime.datetime.utcnow().isoformat().split("T")[1]
        return attr_val

    if attr_type.type == AttrType.URI_WEB:
        attr_val = f"https://sub.domain.com/page-{math.ceil(random.random() * 10000)}/"
        if attr_type.args["allowed_domains"]:
            if attr_type.args["strict_matching"]:
                domain = "sub.domain.com"
            else:
                domain = "domain.com"
            attr_val = attr_val.replace(
                domain, random.choice(attr_type.args["allowed_domains"])
            )
        return attr_val

    if attr_type.type == AttrType.LITERAL:
        attr_val = random.choice(attr_type.args["literal"])
        return attr_val

    if attr_type.type == AttrType.UNION:
        attr_val = generate_attr_val(attr_type=random.choice(attr_type.args["union"]))
        return attr_val

    raise Exception(f"Unknown generator attr '{attr_type}'")



import logging
import os
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import argparse

logger = logging.getLogger("nawah")

NAME_PATTERN = r"^[a-z][a-z0-9_]+$"


def create(args: "argparse.Namespace"):
    """Provides create command functionality to Nawah CLI"""

    if not re.match(r"^[a-z][a-z0-9_]+$", args.name):
        raise Exception(
            "Value for 'name' CLI Arg is invalid. Name should have only small letters, "
            "numbers, and underscores."
        )

    component = args.component
    path = os.path.realpath(args.path)

    logger.info(
        "Checking possibility to create component '%s', at '%s'", component, path
    )

    if os.path.exists(directory := os.path.join(path, args.name)):
        raise Exception(f"Path '{path}' already has item '{args.name}'")

    os.mkdir(directory)
    callables = {
        "app": create_app,
    }
    callables[component](directory, args.name, args.standalone)


def create_app(directory: str, name: str, _):
    with open(os.path.join(directory, "__init__.py"), "w", encoding="UTF-8") as f:
        pass



"""Provides 'ClientApp' dataclass"""

from dataclasses import dataclass
from typing import Literal, MutableSequence, Optional


@dataclass(kw_only=True)
class ClientApp:
    """ClientApp dataclass serves role of defining item for 'client_apps' Config Attr"""

    name: str
    type: Literal["web", "ios", "android"]
    origin: Optional[MutableSequence[str]] = None
    hash: Optional[str] = None



"""Provides 'core' Nawah Package"""

import os

from nawah.classes import Package

from .base import base
from .group import group
from .session import session
from .settings import settings
from .user import user

__version__ = "0.0.0"
with open(
    os.path.join(os.path.dirname(__file__), "..", "..", "version.txt"),
    encoding="UTF-8",
) as f:
    __version__ = f.read().strip()

core = Package(
    name="core",
    api_level="2.0",
    version=__version__,
    modules=[base, user, group, session, settings],
)



"""Provides 'read' Base Function callable"""

from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.classes import Attr
from nawah.config import Config

from .exceptions import UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahSession, Results

DIFF_ATTRS = {
    "user": Attr.ID(),
    "doc": Attr.ID(),
    "attrs": Attr.KV_DICT(key=Attr.STR(), val=Attr.ANY()),
}


async def read_diff(
    *,
    module_name,
    session: "NawahSession",
    query: "Query",
) -> "Results":
    """Reads diff docs for module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="read_diff"
        )

    results = await Data.read(
        session=session,
        collection_name=f"{module.collection}__diff",
        attrs=DIFF_ATTRS,
        query=query,
        extn_attrs=[],
    )

    return {"status": 200, "msg": f'Found {results["count"]} docs', "args": results}



"""Provides Exception classes for 'session' Nawah Module"""

from typing import Literal


class InvalidCredentialsException(Exception):
    """Raised by callable of 'session/auth' endpoint if provided credentials are invalid"""

    status = 403

    def __init__(self):
        super().__init__("Wrong auth credentials")


class InvalidUserException(Exception):
    """Raised by callable of 'session/auth' endpoint if credentials point to a user who is
    'banned', 'deleted', or with 'disabled_password'"""

    status = 403

    def __init__(self, *, reason: Literal["banned", "deleted", "disabled_password"]):
        if reason == "disabled_password":
            msg = "User password is disabled"
        else:
            msg = f"User is {reason}"

        super().__init__(msg)


class AnonReauthException(Exception):
    """Raised by callable of 'session/reauth' endpoint if attempting to reauth ANON user"""

    status = 400

    def __init__(self):
        super().__init__("Reauth is not allowed for ANON user")


class InvalidSessionException(Exception):
    """Raised by callable of 'session/reauth' endpoint if attempting to reauth with invalid
    session _id, or session token, or by callable of 'session/signout' endpoint of attempting
    to sign-out with invalid session"""

    status = 403

    def __init__(self):
        super().__init__("Invalid session")


class ExpiredSessionException(Exception):
    """Raised by callable of 'session/reauth' endpoint if attempting to reauth with expired
    session"""

    status = 403

    def __init__(self):
        super().__init__("Session had expired")


class AnonSignoutException(Exception):
    """Raised by callable of 'session/signout' endpoint if attempting to sign-out ANON user"""

    status = 400

    def __init__(self):
        super().__init__("Sign-out is not allowed for ANON user")



"""Provides 'Query' and related classes and types"""

import copy
import logging
import re
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Union, cast

from bson import ObjectId

from nawah.exceptions import (InvalidQueryOperTypeException,
                              InvalidQueryStepAttrLenException,
                              InvalidQueryStepAttrTypeException,
                              InvalidQueryStepLenException,
                              InvalidQueryStepTypeException,
                              UnknownQueryOperException)

from ._jsonpath import JSONPath
from ._var import Var

if TYPE_CHECKING:
    from nawah.types import (NawahQueryOperAnd, NawahQueryOperOr,
                             NawahQuerySpecial, NawahQueryStep)

logger = logging.getLogger("nawah")

SPECIAL_ATTRS = [
    "$search",
    "$sort",
    "$skip",
    "$limit",
    "$extn",
    "$soft",
    "$attrs",
    "$group",
    "$geo_near",
    "$deleted",
]


QUERY_OPERS = {
    "$eq": (
        type(None),
        str,
        int,
        float,
        list,
        dict,
        ObjectId,
    ),
    "$ne": (
        type(None),
        str,
        int,
        float,
        list,
        dict,
        ObjectId,
    ),
    "$gt": (
        str,
        int,
        float,
    ),
    "$gte": (
        str,
        int,
        float,
    ),
    "$lt": (
        str,
        int,
        float,
    ),
    "$lte": (
        str,
        int,
        float,
    ),
    "$all": (list,),
    "$in": (list,),
    "$nin": (list,),
    "$regex": (str, re.Pattern,),
}


class QueryBase:
    """Serves as abstracted shared parts of Query, QueryMod classes. Meant for internal use only"""

    # pylint: disable=too-few-public-methods

    __pipe: list[Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"]]
    special: "NawahQuerySpecial"

    @property
    def pipe(
        self,
    ) -> tuple[Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"], ...]:
        """Exposes private __pipe attribute as tuple to prevent manipulating pipe directly, and
        assure index is reflecting correct values"""

        return tuple(self.__pipe)

    def __init__(
        self,
        pipe: list[
            Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"]
        ] = None,
        # [TODO] Uncomment when change all current uses to match the API
        # /,
        *,
        special: "NawahQuerySpecial" = None,
    ):
        if not pipe:
            pipe = []
        if not isinstance(pipe, list):
            raise Exception(
                f"Value for 'pipe' should be of type 'list'. Got '{type(pipe)}' "
                "instead"
            )

        if not special:
            special = {}
        if not isinstance(special, dict):
            raise Exception(
                f"Value for 'special' should be of type 'dict'. Got '{type(special)}' "
                "instead"
            )

        self.__pipe = pipe
        self.special = special

        self.__post_init__()

    def __post_init__(self):
        """Emulating a dataclass, QueryBase would call this method after initialisation. Meant to
        be overridden in inheriting class"""

        raise NotImplementedError()


class QueryMod(QueryBase):
    """Serves as value for Func objects 'query_mod' attribute. This is regular Query object,
    excepts it allows use of Var objects as values"""

    # pylint: disable=too-few-public-methods

    def __post_init__(self):
        for step in self.pipe:
            _validate_query_step(step, allow_var=True)


class Query(QueryBase):
    """Serves as value for 'call' Utility 'query' attribute. Query object provides strong-typing
    and strong-validation for the provided values, so it can be detected as early as possible"""

    index: MappingProxyType[str, tuple[tuple[str, Any], ...]]

    @property
    def __pipe(self):
        return self._QueryBase__pipe  # pylint: disable=no-member

    def __post_init__(self):
        for step in self.pipe:
            _validate_query_step(step, allow_var=False)

        self.reset_index()

    def reset_index(self):
        """Resets, and re-creates index for Query object. Meant to be used internally only"""

        index = {}
        _index_query_pipe(pipe=self.pipe, special=self.special, index=index)
        self.index = _freeze_query_index(index)

    def __contains__(self, k):
        if k[0] == "$":
            return k in self.special

        if ":" not in k:
            logger.warning(
                "Query.__contains__ is called with deprecated format: %s. Implicitly, appending "
                "':$eq'",
                k,
            )
            k = f"{k}:$eq"

        return k in self.index

    def __getitem__(self, k):
        if k[0] == "$":
            return self.special[k]

        if ":" not in k:
            logger.warning(
                "Query.__getitem__ is called with deprecated format: %s. Implicitly, appending "
                "':$eq'",
                k,
            )
            k = f"{k}:$eq"

        return QueryItem(self.index[k], self, ":*" in k)

    def __setitem__(self, k, v):
        if k[0] != "$":
            raise Exception("Query.__setitem__ used against non Query Special Attr")

        self.special[k] = v

    def __delitem__(self, k):
        if k[0] != "$":
            raise Exception("Query.__delitem__ used against non Query Special Attr")

        del self.special[k]

    def get(self, pattern: str, /, *, match_oper: bool) -> tuple[str, ...]:
        """Returns tuple with all attrs in Query matching 'pattern'. Value for 'match_oper' would
        determine whether matching step would remove Query Oper when matching or not"""

        query_items = []
        for attr in self.index:
            attr_match = attr

            if not match_oper:
                attr_match = attr_match.split(":")[0]

            if re.match(pattern, attr_match):
                query_items.append(attr)

        return tuple(query_items)

    def append(
        self,
        o: Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"],
        /,
    ):
        """Appends new Query step to current pipe. Before adding step, it will be validated, then
        'reset_index' would be called to re-index object"""

        _validate_query_step(o, allow_var=False)
        self.__pipe.append(o)
        self.reset_index()

    def __deepcopy__(self, memo):
        return Query(copy.deepcopy(self.__pipe), special=copy.deepcopy(self.special))


class QueryItem(list):
    """QueryItem is used to represent Query attrs. Instance of QueryItem is returned from'''
    'Query.__getitem__', which keeps track of the items matching attr in Query. It allows updating
    Query attr values without causing its index to be corrupted"""

    def __init__(self, item_list, query, full_repr):
        self._item_list = item_list
        self._query = query
        super().__init__(
            [value if full_repr else value[list(value)[0]] for path, value in item_list]
        )

    def __setitem__(self, i, o):
        iterable = [i]
        if i == "*":
            iterable = range(len(self))

        for ii in iterable:
            self._item_list[ii][1][list(self._item_list[ii][1])[0]] = o

    def __delitem__(self, i):
        iterable = [i]
        if i == "*":
            iterable = range(len(self))

        for ii in iterable:
            item_path = self._item_list[ii][0].split(".")
            item_parent_path = ".".join(item_path[:-1])
            item_parent = JSONPath(item_parent_path).parse(self._query.pipe)
            del item_parent[int(item_path[-1])]

        # Reset query index, then retrieve value for index to force re-index
        self._query.reset_index()
        self._query.index  # pylint:disable=pointless-statement


def _validate_query_step(
    step: Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"],
    /,
    allow_var: bool,
):
    """Validates value of provided \'Query\' step per \'Query Arg Oper\'. If failed, raises
    \'InvalidQueryArgException\' or \'UnknownQueryArgException\' if \'Query Arg Oper\' is unknow"""

    if not isinstance(step, dict):
        raise InvalidQueryStepTypeException(step_type=type(step))

    if len(step) != 1:
        raise InvalidQueryStepLenException(step_items=tuple(step))

    step_key = list(step)[0]

    if not isinstance(step_key, str):
        # [TODO] Custom Exception
        raise Exception()

    step_cond: Any = step[step_key]  # type: ignore

    if step_key in ("$and", "$or"):
        if not isinstance(step_cond, list):
            # [TODO] Custom exception
            raise Exception()

        for child_step in step_cond:
            _validate_query_step(child_step, allow_var=allow_var)

        return

    if not isinstance(step_cond, dict):
        raise InvalidQueryStepAttrTypeException(step_attr_type=type(step_cond))

    if len(step_cond) > 1:
        raise InvalidQueryStepAttrLenException(step_attr_items=tuple(step_cond))

    step_oper = list(step_cond)[0]

    if step_oper not in QUERY_OPERS:
        raise UnknownQueryOperException(
            attr_name=step_key,
            attr_oper=step_oper,
        )

    step_val = step_cond[step_oper]
    _validate_query_step_val(
        step=step,
        step_key=step_key,
        step_oper=step_oper,
        step_val=step_val,
        allow_var=allow_var,
    )


def _validate_query_step_val(*, step, step_key, step_oper, step_val, allow_var):
    query_oper_types = QUERY_OPERS[step_oper]
    if allow_var:
        query_oper_types = (Var, *QUERY_OPERS[step_oper])

    if not isinstance(step_val, query_oper_types):
        raise InvalidQueryOperTypeException(
            attr_name=step_key,
            attr_oper=step_oper,
            attr_type=QUERY_OPERS[step_oper],
            attr_val=step_val,
        )

    # For step with $regex oper, convert to re.Pattern object. Flags are passed in value of $regex
    # step such if the value is a regexp object. Currently only i (IGNORECASE), m (MULTLINE) are
    # supported. If any other flag is passed it would break the matching sequence, resulting in
    # making the whole value as regexp obejct patter, such as passing '/pattern/.' would be compiled
    # into 're.compile('/pattern/.')', while '/pattern/i' would be compiled into
    # 're.compile('pattern', (re.I, re.M))'
    if step_oper == '$regex':
        if isinstance(step_val, str):
            flags = 0
            if (match := re.match(r'^/(.+)/([mi]{0,2})$', step_val)):
                step_val = match[1]
                if 'i' in match[2]:
                    flags |= re.IGNORECASE
                if 'm' in match[2]:
                    flags |= re.MULTILINE
            step[step_key][step_oper] = re.compile(step_val, flags)

    if isinstance(step_val, list):
        for i, child_step_val in enumerate(step_val):
            # User Query Operator $eq to test against all allowed values for list items
            # IMPORTANT: Appending index to step_key is to prevent nested call from checking
            # step_key again and finding the value to be _id and trying to update step_val
            _validate_query_step_val(
                step=step,
                step_key=f"{step_key}.{i}",
                step_oper="$eq",
                step_val=child_step_val,
                allow_var=allow_var,
            )

    if isinstance(step_val, dict):
        for child_step_attr, child_step_val in step_val.items():
            if not isinstance(child_step_attr, str):
                # [TODO] Custom exception
                raise Exception()

            # User Query Operator $eq to test against all allowed values for dict values
            # IMPORTANT: Appending item key to step_key is to prevent nested call from checking
            # step_key again and finding the value to be _id and trying to update step_val
            _validate_query_step_val(
                step=step,
                step_key=f"{step_key}.{child_step_attr}",
                step_oper="$eq",
                step_val=child_step_val,
                allow_var=allow_var,
            )

    if step_key == "_id":
        if allow_var and isinstance(step_val, Var):
            return

        if isinstance(step_val, list):
            step[step_key][step_oper] = [
                ObjectId(child_step_val) for child_step_val in step_val
            ]
            return

        if not isinstance(step_val, (str, ObjectId)):
            # [TODO] Custom exception
            raise Exception()

        step[step_key][step_oper] = ObjectId(step_val)


def _index_query_pipe(
    *,
    pipe: list[Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"]],
    special: "NawahQuerySpecial",
    index=dict[str, list[tuple[str, Any]]],
    path="",
):
    if not pipe:
        pipe = []

    if not special:
        special = {}

    for i, step in enumerate(pipe):
        _validate_query_step(step, allow_var=False)

        step_key = list(step)[0]

        if step_key in ("$or", "$and"):
            _index_query_pipe(
                pipe=cast(Any, step)[step_key],
                special=special,
                index=index,
                path=f"{path}{i}.",
            )
            continue

        step_cond = step[step_key]  # type: ignore
        step_oper = list(step_cond)[0]
        wild_attr_name = f"{step_key}:*"
        query_attr_name = f"{step_key}:{step_oper}"
        if wild_attr_name not in index:
            index[wild_attr_name] = []
        if query_attr_name not in index:
            index[query_attr_name] = []
        index[wild_attr_name].append((f"{path}{i}", step_cond))
        index[query_attr_name].append((f"{path}{i}", step_cond))


def _freeze_query_index(
    index: dict[str, list[tuple[str, Any]]], /
) -> MappingProxyType[str, tuple[tuple[str, Any], ...]]:
    return MappingProxyType({k: tuple(v) for k, v in index.items()})
